from pathlib import Path
import pprint

from kilosort.gui.logger import XStream
from qtpy import QtCore, QtGui, QtWidgets


class MessageLogBox(QtWidgets.QGroupBox):
    def __init__(self, parent):
        QtWidgets.QGroupBox.__init__(self, parent=parent)
        self.setTitle("Message Log Box")
        self.gui = parent
        self.layout = QtWidgets.QGridLayout()
        self.setLayout(self.layout)
        self.log_box = QtWidgets.QPlainTextEdit()
        self.log_box.setReadOnly(True)
        self.log_box.setLineWrapMode(QtWidgets.QPlainTextEdit.WidgetWidth) 
        self.log_box.setFont(QtGui.QFont("Monospace"))
        self.layout.addWidget(self.log_box, 0, 0, 1, 3)

        self.popout_button = QtWidgets.QPushButton('Show More')
        self.popout_button.clicked.connect(self.show_log_popout)
        self.layout.addWidget(self.popout_button, 1, 0, 1, 1)
        self.popout_window = ExpandedLog()

        self.dump_settings_button = QtWidgets.QPushButton('Dump Settings')
        self.dump_settings_button.clicked.connect(self.dump_settings)
        self.layout.addWidget(self.dump_settings_button, 1, 1, 1, 1)

        self.dump_probe_button = QtWidgets.QPushButton('Dump Probe')
        self.dump_probe_button.clicked.connect(self.dump_probe)
        self.layout.addWidget(self.dump_probe_button, 1, 2, 1, 1)

        log_box_document = self.log_box.document()
        default_font = log_box_document.defaultFont()
        default_font.setPointSize(8)
        log_box_document.setDefaultFont(default_font)

        XStream.stdout().messageWritten.connect(self.update_text)
        XStream.stderr().messageWritten.connect(self.update_text)


    @QtCore.Slot(str)
    def update_text(self, text):
        self.log_box.moveCursor(QtGui.QTextCursor.End)
        self.log_box.appendPlainText(text)
        self.log_box.ensureCursorVisible()
        self.popout_window.update_text(text)

    @QtCore.Slot()
    def show_log_popout(self):
        self.popout_window.show()

    @QtCore.Slot()
    def dump_settings(self):
        # For debugging purposes, check for mismatch between displayed parameter
        # values and the values that are actually being used.
        settings_text = "settings = "
        settings = self.gui.settings_box.settings.copy()
        settings['probe'] = '... (use dump probe)'
        s = pprint.pformat(settings, indent=4, sort_dicts=False)
        settings_text += s[0] + '\n ' + s[1:-1] + '\n' + s[-1]

        self.update_text(settings_text)

    @QtCore.Slot()
    def dump_probe(self):
        # For debugging purposes, make sure probe is loaded correctly.
        probe_text = "probe = "
        probe = self.gui.settings_box.settings['probe']
        p = pprint.pformat(probe, indent=4, sort_dicts=False)
        # insert `np.` so that text can be copied directly to code
        p = 'np.array'.join(p.split('array'))
        p = 'dtype=np.'.join(p.split('dtype='))
        probe_text += p[0] + '\n ' + p[1:-1] + '\n' + p[-1]

        self.update_text(probe_text)

    def prepare_for_new_context(self):
        self.save_log_file()
        self.log_box.clear()

    def save_log_file(self):
        context = self.get_context()
        if context is not None:
            context_path = context.context_path
            log_file_name = context.data_path.stem + ".log"

            log_file_path = Path(context_path) / log_file_name

            if log_file_path.exists():
                with open(log_file_path, "w") as log_file:
                    log_file.write(self.log_box.toPlainText())

    def get_context(self):
        return self.gui.context

    def reset(self):
        self.log_box.clear()


class ExpandedLog(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        self.layout = QtWidgets.QGridLayout()
        self.setLayout(self.layout)
        self.log_box = QtWidgets.QPlainTextEdit()
        self.log_box.setReadOnly(True)
        self.log_box.setLineWrapMode(QtWidgets.QPlainTextEdit.WidgetWidth) 
        self.log_box.setFont(QtGui.QFont("Monospace"))
        self.layout.addWidget(self.log_box, 0, 0, 1, 1)
        
        center = QtWidgets.QApplication.screens()[0].availableGeometry().center()
        self.setGeometry(0, 0, 800, 500)
        geo = self.frameGeometry()
        geo.moveCenter(center)
        self.move(geo.topLeft())

    def update_text(self, text):
        self.log_box.moveCursor(QtGui.QTextCursor.End)
        self.log_box.appendPlainText(text)
        self.log_box.ensureCursorVisible()



import numpy as np

from kilosort.clustering_qr import x_centers
from kilosort.io import load_probe
from kilosort.utils import PROBE_DIR


def random_np2(n_chans=384, n_shanks=4):
    # Generates a probe containing *all* neuropixels 2 contact positions,
    # then randomly subsamples from those positions to get a probe layout
    # corresponding to 384-channel output data.

    probe = {}
    # 12um square contacts with 32um lateral spacing,
    # 15um vertical spacing,
    # 1280 contacts per shank

    # Want alternating 6um, 38um for lateral positions
    xc0 = np.empty(1280)
    xc0[::2] = 6
    xc0[1::2] = 38
    # Then add 250um for each additional shank
    xc = np.concatenate([xc0 + (250*i) for i in range(4)])

    # For vertical positions, start at 6 and increase by 15
    yc0 = (np.arange(640)*15) + 6
    # Each position appears twice (two columns on each shank)
    yc0 = np.repeat(yc0, 2)
    yc = np.concatenate([yc0 for i in range(4)])

    # Repeat 0 1280 times, then repeat 1 1280 times, etc
    kcoords = np.repeat(np.arange(4), 1280)

    # Pick n_chans out of n_shanks
    shanks_used = np.random.choice(range(4), n_shanks, replace=False)
    shank_indices = np.argwhere(np.isin(kcoords, shanks_used))[:,0]
    contact_indices = np.random.choice(shank_indices, n_chans, replace=False)

    return {'xc': xc[contact_indices], 'yc': yc[contact_indices]}


class TestCenters:
    ops = {'dminx': 32}

    def __init__(self, data_directory):
        # This is just here to make sure probes are downloaded before these
        # tests are run.
        pass

    def test_linear(self):
        self.ops['probe'] = load_probe(PROBE_DIR/'Linear16x1_kilosortChanMap.mat')
        centers = x_centers(self.ops)
        # X positions are all 1um
        assert len(centers) == 1
        assert np.abs(centers[0] - 1) < 5

    def test_np1(self):
        self.ops['probe'] = load_probe(PROBE_DIR/'neuropixPhase3B1_kilosortChanMap.mat')
        centers = x_centers(self.ops)
        # One shank from 11um to 59um, should be 1 center near 35um
        assert len(centers) == 1
        assert np.abs(centers[0] - 35) < 5

    def test_np2_1shank(self):
        self.ops['probe'] = load_probe(PROBE_DIR/'NP2_kilosortChanMap.mat')
        centers = x_centers(self.ops)
        # One shank from 0 to 32um, should be 1 center near 16um
        assert len(centers) == 1
        assert np.abs(centers[0] - 16) < 5

    def test_np2_3shank(self):
        self.ops['probe'] = random_np2(n_shanks=3)
        centers = x_centers(self.ops)
        assert len(centers == 3)
        true = np.array([22, 272, 522, 772])
        for c in centers:
            # Each center is within 2 microns of exactly one true center
            print(f'center: {c}')
            assert (np.abs(c - true) < 5).sum() == 1

    def test_np2_4shank(self):
        self.ops['probe'] = random_np2(n_shanks=4)
        centers = x_centers(self.ops)
        # All centers should be within 2 microns of the true values
        print(f'centers: {centers}')
        assert np.allclose(np.sort(centers), np.sort([22, 272, 522, 772]), atol=5)



import numpy as np
import pyqtgraph as pg
from kilosort.gui.logger import setup_logger
from kilosort.gui.palettes import COLORMAP_COLORS
from qtpy import QtCore, QtWidgets

logger = setup_logger(__name__)


class DataViewBox(QtWidgets.QGroupBox):
    channelChanged = QtCore.Signal(int, int)
    modeChanged = QtCore.Signal(str, int)
    updateContext = QtCore.Signal(object)
    intervalUpdated = QtCore.Signal()

    def __init__(self, parent):
        QtWidgets.QGroupBox.__init__(self, parent=parent)

        self.gui = parent

        self.data_view_widget = KSPlotWidget(useOpenGL=True)
        self.data_x_axis = self.data_view_widget.getAxis("bottom")
        self.plot_item = self.data_view_widget.getPlotItem()
        self.data_view_box = self.data_view_widget.getViewBox()
        self.colormap_image = None

        self.data_seek_widget = pg.PlotWidget(useOpenGL=True)
        self.seek_view_box = self.data_seek_widget.getViewBox()
        self.time_seek = pg.InfiniteLine(
            pen=pg.mkPen((255, 0, 0, 128), width=3), movable=True, name="indicator"
        )
        self.time_label = pg.TextItem(color=(180, 180, 180))

        self.raw_button = QtWidgets.QPushButton("Raw")
        self.whitened_button = QtWidgets.QPushButton("Whitened")
        # Set minimum and maximum time in seconds for spike sorting.
        # This should truncate the data view if specified.
        self.tmin_input = self.gui.settings_box.tmin_input
        self.tmin = float(self.tmin_input.text())
        self.tmax_input = self.gui.settings_box.tmax_input
        self.tmax = float(self.tmax_input.text())

        self.mode_buttons_group = QtWidgets.QButtonGroup(self)
        self.view_buttons_group = QtWidgets.QButtonGroup(self)

        self.view_buttons = {
            "raw": self.raw_button,
            "whitened": self.whitened_button,
        }

        self.input_fields = [self.tmin_input, self.tmax_input]

        self._view_button_checked_bg_colors = {
            "raw": "white",
            "whitened": "lightblue",
        }

        self._keys = [
            "raw",
            "whitened",
        ]

        self.view_buttons_state = {key: False for key in self._keys}

        self.primary_channel = 0
        self.current_time = 0
        self.plot_range = 0.1  # seconds

        self.highpass_filter = None

        self.whitening_matrix = None
        self.whitened_traces = None
        self.prediction_traces = None
        self.residual_traces = None

        self.sorting_status = {
            "preprocess": False,
            "spikesort": False,
            "export": False,
        }

        # traces settings
        self.traces_plot_items = {
            "raw": [],
            "whitened": [],
            "prediction": [],
            "residual": [],
        }
        self.good_channel_color = (255, 255, 255)
        self.bad_channel_color = (100, 100, 100)
        self.channels_displayed_traces = 32
        self.channels_displayed_colormap = None
        self.seek_range = (0, 100)
        self.scale_factor = 1.0
        self.traces_scaling_factor = {
            "raw": 1,
            "whitened": 15,
        }
        self.traces_curve_color = {
            "raw": "w",
            "whitened": "c",
        }

        # colormap settings
        self._colors = COLORMAP_COLORS

        self.colormap_min = 0.0
        self.colormap_max = 1.0
        self.lookup_table = self.generate_lookup_table(
            self.colormap_min, self.colormap_max
        )

        self.thread_pool = QtCore.QThreadPool()

        self.setup()

    def setup(self):
        self.setTitle("Data View")

        layout = QtWidgets.QVBoxLayout()

        data_view_layout = QtWidgets.QHBoxLayout()
        data_view_layout.addWidget(self.data_view_widget)

        self.time_label.setParentItem(self.seek_view_box)
        self.time_label.setPos(0, 0)
        self.data_seek_widget.addItem(self.time_seek)

        self.time_seek.sigPositionChanged.connect(self.update_seek_text)
        self.time_seek.sigPositionChanged.connect(self.update_seek_position)

        self.data_view_widget.setMenuEnabled(False)
        self.data_view_widget.setMouseEnabled(True)
        self.data_view_widget.mouseEnabled = True
        self.data_view_widget.hideAxis("left")
        self.data_view_widget.disableAutoRange()
        self.data_view_widget.sceneObj.sigMouseClicked.connect(self.on_scene_clicked)

        self.data_view_widget.signalChangeChannel.connect(
            self.on_wheel_scroll_plus_control
        )
        self.data_view_widget.signalChangeTimeRange.connect(
            self.on_wheel_scroll_plus_shift
        )
        self.data_view_widget.signalChangeScaling.connect(self.on_wheel_scroll_plus_alt)
        self.data_view_widget.signalChangeTimePoint.connect(self.on_wheel_scroll)

        self.data_seek_widget.setMenuEnabled(False)
        self.data_seek_widget.setMouseEnabled(False, False)
        self.data_seek_widget.hideAxis("left")
        self.data_seek_widget.sceneObj.sigMouseClicked.connect(self.seek_clicked)

        data_controls_layout = QtWidgets.QHBoxLayout()


        for key in self._keys:
            button = self.view_buttons[key]
            button.setCheckable(True)
            button.setStyleSheet("QPushButton {background-color: black; color: white;}")
            button.clicked.connect(self.on_views_clicked)
            self.view_buttons_group.addButton(button)

        self.raw_button.setChecked(True)
        self.raw_button.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )

        self.enable_view_buttons()

        self.view_buttons_group.setExclusive(True)

        self.view_buttons_state = [
            self.view_buttons[key].isChecked() for key in self._keys
        ]


        # Connect time controls
        self.tmin_input.editingFinished.connect(self.on_tmin_edited)
        self.tmax_input.editingFinished.connect(self.on_tmax_edited)

        data_controls_layout.addStretch(0)
        data_controls_layout.addWidget(self.raw_button)
        data_controls_layout.addWidget(self.whitened_button)

        data_seek_layout = QtWidgets.QHBoxLayout()
        data_seek_layout.addWidget(self.data_seek_widget)

        layout.addLayout(data_view_layout, 85)
        layout.addLayout(data_controls_layout, 3)
        layout.addLayout(data_seek_layout, 10)

        self.setLayout(layout)

    @QtCore.Slot()
    def on_tmin_edited(self):
        try:
            self.tmin = float(self.tmin_input.text())
            self.check_time_interval()
        except ValueError:
            logger.exception('Could not convert tmin to float.')
        except AssertionError:
            logger.exception('Invalid tmin,tmax: must have 0 <= tmin < tmax')

    @QtCore.Slot()
    def on_tmax_edited(self):
        try:
            self.tmax = float(self.tmax_input.text())
            self.check_time_interval()
        except ValueError:
            logger.exception('Could not convert tmax to float.')
        except AssertionError:
            logger.exception('Invalid tmin,tmax: must have 0 <= tmin < tmax')

    def check_time_interval(self):
        # TODO: any other checks needed?
        assert self.tmin >= 0
        assert self.tmin < self.tmax

    @QtCore.Slot()
    def on_views_clicked(self):
        current_state = {key: self.view_buttons[key].isChecked() for key in self._keys}

        if current_state != self.view_buttons_state:
            self.view_buttons_state = current_state

            for view, state in self.view_buttons_state.items():
                button = self.view_buttons[view]
                if state:
                    color = self._view_button_checked_bg_colors[view]
                    button.setStyleSheet(
                        f"QPushButton {{background-color: {color}; color: black;}}"
                    )
                else:
                    if button.isEnabled():
                        button.setStyleSheet(
                            "QPushButton {background-color: black; color: white;}"
                        )
                    else:
                        button.setStyleSheet(
                            "QPushButton {background-color: black; color: gray;}"
                        )

            self.update_plot()

    @QtCore.Slot(float)
    def on_wheel_scroll(self, direction):
        if self.context_set():
            self.shift_current_time(direction)

    @QtCore.Slot(float)
    def on_wheel_scroll_plus_control(self, direction):
        if self.context_set():
            self.change_displayed_channel_count(direction)

    @QtCore.Slot(float)
    def on_wheel_scroll_plus_shift(self, direction):
        if self.context_set():
            self.change_plot_range(direction)

    @QtCore.Slot(float)
    def on_wheel_scroll_plus_alt(self, direction):
        if self.context_set():
            self.change_plot_scaling(direction)

    def context_set(self):
        return self.get_context() is not None

    def get_context(self):
        return self.gui.context

    def change_primary_channel(self, channel):
        self.primary_channel = channel
        self.channelChanged.emit(self.primary_channel, self.get_currently_displayed_channel_count())
        self.update_plot()

    def shift_primary_channel(self, shift):
        primary_channel = self.primary_channel
        primary_channel += shift * 5
        total_channels = self.get_total_channels()
        if (0 <= primary_channel < total_channels) and total_channels is not None:
            self.primary_channel = primary_channel
            self.channelChanged.emit(self.primary_channel, self.get_currently_displayed_channel_count())
            self.update_plot()

    def get_currently_displayed_channel_count(self):
        count = self.channels_displayed_colormap
        if count is None:
            count = self.get_total_channels()
        return count

    def set_currently_displayed_channel_count(self, count):
        self.channels_displayed_colormap = count

    def get_total_channels(self):
        return self.gui.probe_view_box.total_channels

    def change_displayed_channel_count(self, direction):
        current_channel = self.primary_channel
        total_channels = self.get_total_channels()
        current_count = self.get_currently_displayed_channel_count()

        new_count = current_count + (direction * 5)
        if (current_channel + new_count) <= total_channels:
            self.set_currently_displayed_channel_count(new_count)
            self.channelChanged.emit(self.primary_channel, new_count)
            self.refresh_plot_on_displayed_channel_count_change()
        elif new_count <= 0 and current_count != 1:
            self.set_currently_displayed_channel_count(1)
            self.channelChanged.emit(self.primary_channel, 1)
            self.refresh_plot_on_displayed_channel_count_change()
        elif (current_channel + new_count) > total_channels:
            self.set_currently_displayed_channel_count(total_channels - current_channel)
            self.channelChanged.emit(self.primary_channel, total_channels)
            self.refresh_plot_on_displayed_channel_count_change()

    def refresh_plot_on_displayed_channel_count_change(self):
        self.plot_item.clear()
        self.update_plot()

    def shift_current_time(self, direction):
        time_shift = direction * self.plot_range / 4  # seconds
        current_time = self.current_time
        new_time = current_time + time_shift
        seek_range_min = self.seek_range[0]
        seek_range_max = self.seek_range[1] - self.plot_range
        if seek_range_min <= new_time <= seek_range_max:
            # if new time is in acceptable limits
            self.time_seek.setPos(new_time)
        elif new_time <= seek_range_min:
            # if new time exceeds lower bound of data
            self.time_seek.setPos(seek_range_min)
        elif new_time >= seek_range_max:
            # if new time exceeds upper bound of data
            self.time_seek.setPos(seek_range_max)

    def change_plot_range(self, direction):
        current_range = self.plot_range
        # neg sign to reverse scrolling behaviour
        new_range = current_range * (1.2 ** -direction)
        if 0.005 < new_range < 1.0:
            diff_range = new_range - current_range
            current_time = self.current_time
            new_time = current_time - diff_range/2.
            seek_range_min = self.seek_range[0]
            seek_range_max = self.seek_range[1]
            if new_time < seek_range_min:
                # if range exceeds lower limit on zooming
                # set lower limit as current time
                self.plot_range = new_range
                self.time_seek.setPos(seek_range_min)
            elif new_time + new_range > seek_range_max:
                # if range exceeds upper limit on zooming
                # set (upper limit - new time) as current range
                alt_range = seek_range_max - new_time
                self.plot_range = alt_range
                self.time_seek.setPos(new_time)
            else:
                # if range is acceptable
                self.plot_range = new_range
                self.time_seek.setPos(new_time)

    def change_plot_scaling(self, direction):
        colormap_min = self.colormap_min + (direction * 0.05)
        colormap_max = self.colormap_max - (direction * 0.05)
        if 0.0 <= colormap_min < colormap_max <= 1.0:
            self.colormap_min = colormap_min
            self.colormap_max = colormap_max
            self.lookup_table = self.generate_lookup_table(
                self.colormap_min, self.colormap_max
            )

            self.update_plot()

    def on_scene_clicked(self, ev):
        if self.context_set():
            if ev.button() == QtCore.Qt.LeftButton:
                sample_rate = self.get_context().params["fs"]
                x_pos = self.colormap_image.mapFromScene(ev.pos()).x()
                plot_range = self.plot_range * sample_rate
                fraction = x_pos / plot_range
                if fraction > 0.5:
                    self.shift_current_time(direction=1)
                else:
                    self.shift_current_time(direction=-1)

    def update_context(self, new_context):
        self.updateContext.emit(new_context)

    def seek_clicked(self, ev):
        if self.context_set():
            new_time = self.seek_view_box.mapSceneToView(ev.pos()).x()
            seek_range_min = self.seek_range[0]
            seek_range_max = self.seek_range[1]
            if seek_range_min <= new_time < seek_range_max - self.plot_range:
                self.time_seek.setPos(new_time)
            elif seek_range_max - self.plot_range <= new_time < seek_range_max:
                # make sure the plotted data is always as wide as the plot range
                self.time_seek.setPos(seek_range_max - self.plot_range)

    def setup_seek(self, context):
        binary_file = context.binary_file
        sample_rate = context.params["fs"]

        timepoints = binary_file.shape[0]
        min_time = max(0, self.tmin)
        max_time = (timepoints / sample_rate) + min_time

        self.data_seek_widget.setXRange(
            min=min_time,
            max=max_time,
            padding=0.02
        )
        self.time_seek.setPos(min_time)
        self.time_seek.setBounds((min_time, max_time))
        self.seek_range = (min_time, max_time)

    def update_seek_text(self, seek):
        position = seek.pos()[0]
        self.time_label.setText("t={0:.2f} s".format(position))

    def update_seek_position(self, seek):
        position = seek.pos()[0]
        self.current_time = position - self.tmin
        try:
            self.update_plot()
        except ValueError:
            self.time_seek.setPos(self.seek_range[1] - self.plot_range)

    def change_sorting_status(self, status_dict):
        self.sorting_status = status_dict
        self.enable_view_buttons()

    @QtCore.Slot()
    def enable_view_buttons(self):
        self.raw_button.click()

        if self.whitening_matrix is not None:
            self.whitened_button.setEnabled(True)
            self.whitened_button.setStyleSheet(
                "QPushButton {background-color: black; color: white;}"
            )
        else:
            self.whitened_button.setDisabled(True)
            self.whitened_button.setStyleSheet(
                "QPushButton {background-color: black; color: gray;}"
            )

    def reset(self):
        self.plot_item.clear()
        self.clear_cached_whitening_matrix()

    def prepare_for_new_context(self):
        self.plot_item.clear()
        self.clear_cached_whitening_matrix()

    def clear_cached_whitening_matrix(self):
        self.whitening_matrix = None

    def generate_lookup_table(self, colormap_min, colormap_max, num_points=8192):
        assert colormap_min >= 0.0 and colormap_max <= 1.0
        positions = np.linspace(colormap_min, colormap_max, len(self._colors))
        color_map = pg.ColorMap(pos=positions, color=self._colors)
        return color_map.getLookupTable(nPts=num_points)

    def add_image_to_plot(self, raw_traces, level_min, level_max):
        image_item = pg.ImageItem(setPxMode=False)
        image_item.setImage(
            raw_traces,
            autoLevels=False,
            lut=self.lookup_table,
            levels=(level_min, level_max),
            autoDownsample=False,
        )
        self.colormap_image = image_item
        self.plot_item.addItem(image_item)

    @QtCore.Slot(object)
    def set_whitening_matrix(self, array):
        self.whitening_matrix = array

    @QtCore.Slot(object)
    def set_highpass_filter(self, filter):
        self.highpass_filter = filter

    def update_plot(self, context=None):
        if context is None:
            context = self.gui.context

        if context is not None:  # since context may still be None
            if self.colormap_image is not None:
                self.plot_item.removeItem(self.colormap_image)
                self.colormap_image = None

            params = context.params
            probe = context.probe
            binary_file = context.binary_file
            filt_binary_file = context.filt_binary_file

            sample_rate = params["fs"]

            start_time = int(self.current_time * sample_rate)
            time_range = int(self.plot_range * sample_rate)
            end_time = start_time + time_range

            self.data_view_widget.setXRange(
                min=0,
                max=time_range,
                padding=0.02,
            )
            self.data_view_widget.setLimits(
                xMin=0,
                xMax=time_range
            )

            max_channels = np.size(probe["chanMap"])

            start_channel = self.primary_channel
            active_channels = self.get_currently_displayed_channel_count()
            if active_channels is None:
                active_channels = self.get_total_channels()
            end_channel = start_channel + active_channels

            # good channels after start_channel to display
            to_display = np.arange(start_channel, end_channel, dtype=int)
            to_display = to_display[to_display < max_channels].tolist()

            self._update_colormap(
                params=params,
                probe=probe,
                binary_file=binary_file,
                filt_binary_file=filt_binary_file,
                to_display=to_display,
                start_time=start_time,
                end_time=end_time,
            )

            min_tick = start_time + (self.tmin * sample_rate)
            self.data_x_axis.setTicks(
                [
                    [
                        (pos, f"{(min_tick + pos) / sample_rate:.3f}")
                        for pos in np.linspace(0, time_range, 20)
                    ]
                ]
            )

            self.data_view_widget.autoRange()


    def _update_colormap(
            self,
            params,
            probe,
            binary_file,
            filt_binary_file,
            to_display,
            start_time,
            end_time
    ):

        if self.raw_button.isChecked():
            colormap_min, colormap_max = -32.0, 32.0
            raw_traces = binary_file[start_time:end_time].cpu().numpy()
            self.add_image_to_plot(
                raw_traces[to_display, :].T,
                colormap_min,
                colormap_max,
            )

        elif self.whitened_button.isChecked():
            whitened_traces = filt_binary_file[start_time:end_time].cpu().numpy()

            colormap_min, colormap_max = -4.0, 4.0
            self.add_image_to_plot(
                whitened_traces[to_display, :].T,
                colormap_min,
                colormap_max,
            )


class KSPlotWidget(pg.PlotWidget):
    signalChangeTimePoint = QtCore.Signal(float)
    signalChangeChannel = QtCore.Signal(float)
    signalChangeTimeRange = QtCore.Signal(float)
    signalChangeScaling = QtCore.Signal(float)

    def __init__(self, *args, **kwargs):
        super(KSPlotWidget, self).__init__(*args, **kwargs)

    def wheelEvent(self, ev):
        delta = ev.angleDelta().y()
        if delta == 0:
            delta = ev.angleDelta().x()

        direction = delta / np.abs(delta)
        modifiers = QtWidgets.QApplication.keyboardModifiers()

        if modifiers == QtCore.Qt.ControlModifier:
            # control pressed while scrolling
            self.signalChangeChannel.emit(direction)

        elif modifiers == QtCore.Qt.AltModifier:
            # alt pressed while scrolling
            self.signalChangeScaling.emit(direction)

        elif modifiers == QtCore.Qt.ShiftModifier:
            # shift pressed while scrolling
            self.signalChangeTimeRange.emit(direction)

        else:
            # other key / no key pressed while scrolling
            self.signalChangeTimePoint.emit(direction)

        ev.accept()
        return

    def mouseMoveEvent(self, ev):
        pass



import time

import numpy as np
import torch
from qtpy import QtCore

from kilosort.gui.logger import setup_logger
from kilosort.run_kilosort import (
    initialize_ops, compute_preprocessing, compute_drift_correction,
    detect_spikes, cluster_spikes, save_sorting
    )

logger = setup_logger(__name__)


class KiloSortWorker(QtCore.QThread):
    finishedPreprocess = QtCore.Signal(object)
    finishedSpikesort = QtCore.Signal(object)
    finishedAll = QtCore.Signal(object)
    progress_bar = QtCore.Signal(int)
    plotDataReady = QtCore.Signal(str)

    def __init__(self, context, results_directory, steps,
                 device=torch.device('cuda'), file_object=None, *args, **kwargs):
        super(KiloSortWorker, self).__init__(*args, **kwargs)
        self.context = context
        self.data_path = context.data_path
        self.results_directory = results_directory
        assert isinstance(steps, list) or isinstance(steps, str)
        self.steps = steps if isinstance(steps, list) else [steps]
        self.device = device
        self.file_object = file_object

    def run(self):
        if "spikesort" in self.steps:
            settings = self.context.params
            probe = self.context.probe
            settings["data_dir"] = self.data_path.parent
            settings["filename"] = self.data_path
            results_dir = self.results_directory
            if not results_dir.exists():
                logger.info(f"Results dir at {results_dir} does not exist."
                             "The folder will be created.")
                results_dir.mkdir(parents=True)

            tic0 = time.time()

            # TODO: make these options in GUI
            do_CAR=True
            invert_sign=False
        
            if not do_CAR:
                print("Skipping common average reference.")

            if settings['nt0min'] is None:
                settings['nt0min'] = int(20 * settings['nt']/61)
            data_dtype = settings['data_dtype']
            device = self.device

            ops = initialize_ops(settings, probe, data_dtype, do_CAR,
                                 invert_sign, device)

            # TODO: add support for file object through data conversion
            # Set preprocessing and drift correction parameters
            ops = compute_preprocessing(ops, self.device, tic0=tic0,
                                        file_object=self.file_object)
            np.random.seed(1)
            torch.cuda.manual_seed_all(1)
            torch.random.manual_seed(1)
            ops, bfile, st0 = compute_drift_correction(
                ops, self.device, tic0=tic0, progress_bar=self.progress_bar,
                file_object=self.file_object
                )

            # Will be None if nblocks = 0 (no drift correction)
            if st0 is not None:
                self.dshift = ops['dshift']
                self.st0 = st0
                self.plotDataReady.emit('drift')

            # Sort spikes and save results
            st, tF, Wall0, clu0 = detect_spikes(ops, self.device, bfile, tic0=tic0,
                                                progress_bar=self.progress_bar)

            self.Wall0 = Wall0
            self.wPCA = torch.clone(ops['wPCA'].cpu()).numpy()
            self.clu0 = clu0
            self.plotDataReady.emit('diagnostics')

            clu, Wall = cluster_spikes(st, tF, ops, self.device, bfile, tic0=tic0,
                                       progress_bar=self.progress_bar)
            ops, similar_templates, is_ref, est_contam_rate = \
                save_sorting(ops, results_dir, st, clu, tF, Wall, bfile.imin, tic0)

            self.ops = ops
            self.st = st
            self.clu = clu
            self.tF = tF
            self.is_refractory = is_ref
            self.plotDataReady.emit('probe')

            logger.info(f"Spike sorting output saved in\n{results_dir}")
            self.finishedSpikesort.emit(self.context)



from io import StringIO
import time
import numpy as np
from kilosort import spikedetect, preprocessing, CCG
import torch 
from torch.nn.functional import conv1d, max_pool2d, max_pool1d
from tqdm import tqdm

def prepare_extract(ops, U, nC, device=torch.device('cuda')):
    ds = (ops['xc'] - ops['xc'][:, np.newaxis])**2 +  (ops['yc'] - ops['yc'][:, np.newaxis])**2 
    iCC = np.argsort(ds, 0)[:nC]
    iCC = torch.from_numpy(iCC).to(device)
    iU = torch.argmax((U**2).sum(1), -1)
    Ucc = U[torch.arange(U.shape[0]),:,iCC[:,iU]]
    return iCC, iU, Ucc

def extract(ops, bfile, U, device=torch.device('cuda'), progress_bar=None):
    nC = ops['settings']['nearest_chans']
    iCC, iU, Ucc = prepare_extract(ops, U, nC, device=device)
    ops['iCC'] = iCC
    ops['iU'] = iU

    nt = ops['nt']
    tiwave = torch.arange(-(nt//2), nt//2+1, device=device) 

    ctc = prepare_matching(ops, U)
    st = np.zeros((10**6, 3), 'float64')
    tF  = torch.zeros((10**6, nC , ops['settings']['n_pcs']))

    k = 0

    for ibatch in tqdm(np.arange(bfile.n_batches), miniters=200 if progress_bar else None, 
                        mininterval=60 if progress_bar else None):
        X = bfile.padded_batch_to_torch(ibatch, ops)

        stt, amps, Xres = run_matching(ops, X, U, ctc, device=device)

        xfeat = Xres[iCC[:, iU[stt[:,1:2]]],stt[:,:1] + tiwave] @ ops['wPCA'].T
        xfeat += amps * Ucc[:,stt[:,1]]

        nsp = len(stt)   
        if k+nsp>st.shape[0]:                     
            st = np.concatenate((st, np.zeros_like(st)), 0)
            tF  = torch.cat((tF,  torch.zeros_like(tF)), 0)

        stt = stt.double()
        #st[k:k+nsp,0] = ((stt[:,0]-nt)/ops['fs'] + ibatch * (ops['batch_size']/ops['fs'])).cpu().numpy()
        left_pad = 0 if ibatch == 0 else nt
        st[k:k+nsp,0] = ((stt[:,0]-left_pad) + ibatch * (ops['batch_size'])).cpu().numpy() - nt//2 + ops['nt0min']

        st[k:k+nsp,1] = stt[:,1].cpu().numpy()
        st[k:k+nsp,2] = amps[:,0].cpu().numpy()
        
        tF[k:k+nsp]  = xfeat.transpose(0,1).cpu()#.numpy()

        k+= nsp
        
        if progress_bar is not None:
            progress_bar.emit(int((ibatch+1) / bfile.n_batches * 100))

    isort = np.argsort(st[:k,0])

    st = st[isort]
    tF = tF[isort]

    return st, tF, ops

def align_U(U, ops, device=torch.device('cuda')):
    Uex = torch.einsum('xyz, zt -> xty', U.to(device), ops['wPCA'])
    X = Uex.reshape(-1, ops['Nchan']).T
    X = conv1d(X.unsqueeze(1), ops['wTEMP'].unsqueeze(1), padding=ops['nt']//2)
    Xmax = X.abs().max(0)[0].max(0)[0].reshape(-1, ops['nt'])
    imax = torch.argmax(Xmax, 1)

    Unew = Uex.clone() 
    for j in range(ops['nt']):
        ix = imax==j
        Unew[ix] = torch.roll(Unew[ix], ops['nt']//2 - j, -2)
    Unew = torch.einsum('xty, zt -> xzy', Unew, ops['wPCA'])#.transpose(1,2).cpu()
    return Unew, imax


def postprocess_templates(Wall, ops, clu, st, device=torch.device('cuda')):
    Wall2, _ = align_U(Wall, ops, device=device)
    #Wall3, _= remove_duplicates(ops, Wall2)
    Wall3, _, _ = merging_function(ops, Wall2.transpose(1,2), clu, st[:,0],
                                   0.9, 'mu', device=device)
    Wall3 = Wall3.transpose(1,2).to(device)
    return Wall3

def prepare_matching(ops, U):
    nt = ops['nt']
    W = ops['wPCA'].contiguous()
    WtW = conv1d(W.reshape(-1, 1,nt), W.reshape(-1, 1 ,nt), padding = nt) 
    WtW = torch.flip(WtW, [2,])

    #mu = (U**2).sum(-1).sum(-1)**.5
    #U2 = U / mu.unsqueeze(-1).unsqueeze(-1)

    UtU = torch.einsum('ikl, jml -> ijkm',  U, U)
    ctc = torch.einsum('ijkm, kml -> ijl', UtU, WtW)

    return ctc

def run_matching(ops, X, U, ctc, device=torch.device('cuda')):
    Th = ops['Th_learned']
    nt = ops['nt']
    W = ops['wPCA'].contiguous()

    nm = (U**2).sum(-1).sum(-1)
    #mu = nm**.5 
    #U2 = U / mu.unsqueeze(-1).unsqueeze(-1)

    B = conv1d(X.unsqueeze(1), W.unsqueeze(1), padding=nt//2)
    B = torch.einsum('ijk, kjl -> il', U, B)

    trange = torch.arange(-nt, nt+1, device=device) 
    tiwave = torch.arange(-(nt//2), nt//2+1, device=device) 

    st = torch.zeros((100000,2), dtype = torch.int64, device = device)
    amps = torch.zeros((100000,1), dtype = torch.float, device = device)
    k = 0

    Xres = X.clone()
    lam = 20

    for t in range(100):
        # Cf = 2 * B - nm.unsqueeze(-1) 
        Cf = torch.relu(B)**2 /nm.unsqueeze(-1)
        #a = 1 + lam
        #b = torch.relu(B) + lam * mu.unsqueeze(-1)
        #Cf = b**2 / a - lam * mu.unsqueeze(-1)**2

        Cf[:, :nt] = 0
        Cf[:, -nt:] = 0

        Cfmax, imax = torch.max(Cf, 0)
        Cmax  = max_pool1d(Cfmax.unsqueeze(0).unsqueeze(0), (2*nt+1), stride = 1, padding = (nt))

        #print(Cfmax.shape)
        #import pdb; pdb.set_trace()
        cnd1 = Cmax[0,0] > Th**2
        cnd2 = torch.abs(Cmax[0,0] - Cfmax) < 1e-9
        xs = torch.nonzero(cnd1 * cnd2)

        if len(xs)==0:
            #print('iter %d'%t)
            break

        iX = xs[:,:1]
        iY = imax[iX]

        #isort = torch.sort(iX)

        nsp = len(iX)    
        st[k:k+nsp, 0] = iX[:,0]
        st[k:k+nsp, 1] = iY[:,0]
        amps[k:k+nsp] = B[iY,iX] / nm[iY]
        amp = amps[k:k+nsp]

        k+= nsp

        #amp = B[iY,iX] 

        n = 2
        for j in range(n):
            Xres[:, iX[j::n] + tiwave]  -= amp[j::n] * torch.einsum('ijk, jl -> kil', U[iY[j::n,0]], W)
            B[   :, iX[j::n] + trange]  -= amp[j::n] * ctc[:,iY[j::n,0],:]

    st = st[:k]
    amps = amps[:k]

    return  st, amps, Xres


def merging_function(ops, Wall, clu, st, r_thresh=0.5, mode='ccg', device=torch.device('cuda')):
    clu2 = clu.copy()
    clu_unq, ns = np.unique(clu2, return_counts = True)

    Ww = Wall.to(device)
    NN = len(Ww)

    isort = np.argsort(ns)[::-1]

    is_merged = np.zeros(NN, 'bool')
    is_good = np.zeros(NN,)

    acg_threshold = ops['settings']['acg_threshold']
    ccg_threshold = ops['settings']['ccg_threshold']
    if mode == 'ccg':
        is_ref, est_contam_rate = CCG.refract(clu, st/ops['fs'],
                                              acg_threshold=acg_threshold,
                                              ccg_threshold=ccg_threshold)

    nt = ops['nt']
    W = ops['wPCA'].contiguous()
    WtW = conv1d(W.reshape(-1, 1,nt), W.reshape(-1, 1 ,nt), padding = nt) 
    WtW = torch.flip(WtW, [2,])

    t = 0
    nmerge = 0
    while t<NN:
        #if t%100==0:
            #print(t, nmerge)

        kk = clu_unq[isort[t]]

        if (mode == 'ccg') and is_ref[kk]==0:
            t += 1
            continue

        if is_merged[kk]:            
            t += 1
            continue

        mu = (Ww**2).sum((1,2), keepdims=True)**.5
        Wnorm = Ww / (1e-6 + mu)

        UtU = torch.einsum('lk, jlm -> jkm',  Wnorm[kk], Wnorm)
        ctc = torch.einsum('jkm, kml -> jl', UtU, WtW)

        cmax = ctc.max(1)[0]
        cmax[kk] = 0

        jsort = np.argsort(cmax.cpu().numpy())[::-1]

        if mode == 'ccg':
            st0 = st[clu2==kk] / ops['fs']
        
        is_ccg  = 0
        for j in range(NN):
            jj = jsort[j]
            if cmax[jj] < r_thresh:
                break
            # compare with CCG
            if mode == 'ccg':
                st1 = st[clu2==jj] / ops['fs']
                _, is_ccg, _ = CCG.check_CCG(st0, st1, acg_threshold=acg_threshold,
                                             ccg_threshold=ccg_threshold)        
            else:
                dmu = 2 * (mu[kk] - mu[jj]) / (mu[kk] + mu[jj])
                is_ccg = dmu.abs() < 0.2

            if is_ccg:
                is_merged[jj] = 1
                Ww[kk] = ns[kk]/(ns[kk]+ns[jj]) * Ww[kk] + ns[jj]/(ns[kk]+ns[jj]) * Ww[jj]            
                Ww[jj] = 0

                ns[kk] += ns[jj]
                ns[jj] = 0
                clu2[clu2==jj] = kk            

                break

        if is_ccg==0:            
            t +=1    
        else:                
            nmerge+=1
    
    imap = np.cumsum((~is_merged).astype('int32')) - 1
    if imap.size > 0:
        # Otherwise, everything has been merged into a single cluster
        clu2 = imap[clu2]

    Ww = Ww[~is_merged]

    if mode == 'ccg':
        is_ref = is_ref[~is_merged]
    else:
        is_ref = None

    return Ww.cpu(), clu2, is_ref
